// Shared Array
// Make a user-specified number of increments to random elements of an
// array of user-specified size

#include <assert.h>
#include <magic_services.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include <sys/time.h>
#include <sys/types.h>

#include "barrier.h"

#define MAX_PROCS 16

/* shared variables */
// determine with: /usr/platform/sun4u/sbin/prtdiag -v
int processor_ids[MAX_PROCS] = {0, 1, 4, 5, 8, 9, 12, 13, 16, 17, 20, 21, 24,
                                 25, 28, 29};
int num_elems; // the number of elements in the shared array
int num_incr; // the number of increments to make to the shared array
int num_threads;
int* sa; // shared array
pthread_mutex_t sa_lock; // mutex for whole shared array
barrier_t bar; // barrier

void print_array(int* array, int length) {
  int i;

  printf("[");
  for (i = 0; i < length - 1; i++) {
    printf("%d, ", array[i]);
  }
  printf("%d]\n", array[length - 1]);
}

void verify(int* array, int length, int num_incr) {
  int i; // loop counter
  int sum = 0;

  for (i = 0; i < length; i++) {
    sum += array[i];
  }

  if (sum == num_incr) {
    printf("CORRECT!\n");
  } else {
    printf("ERROR!");
    printf("Expected: %d Counted: %d\n", num_incr, sum);
  }
}

// perform random increments
void* increment(void* my_id) {
  float RAND_FACTOR; // rand multiplier to get an index in the appropriate range
  int i; // loop counter
  int id = (int) my_id;
  hrtime_t start_time, end_time;
  int local_sense = 0; // for sense-reversing barrier

  // bind thread to a processor, ensuring at most one thread per proc
  assert(!processor_bind(P_LWPID, P_MYID, processor_ids[id], NULL));

  RAND_FACTOR = (float) (num_elems - 1) / RAND_MAX;
  //printf("RAND_FACTOR=%f\n", RAND_FACTOR);

  barrier(&bar, &local_sense);

  if (id == 0) {
      start_time = gethrtime();
      RUBY_MAGIC_CALL(Do_Breakpoint, 0);
  }

  barrier(&bar, &local_sense);

  for (i = 0; i < (num_incr / num_threads); i++) {
    /*
    index = (int) (rand() * RAND_FACTOR);
    if (index < 0 || index > num_elems - 1) {
      printf("BAD INDEX: idx=%d\n", index);
    } else {
      sa[index]++;
    }
    */
    pthread_mutex_lock(&sa_lock);
    sa[(int) (rand() * RAND_FACTOR)]++;
    pthread_mutex_unlock(&sa_lock);
  }

  barrier(&bar, &local_sense);

  if (id == 0) {
      RUBY_MAGIC_CALL(Do_Breakpoint, 0);
      end_time = gethrtime();
      printf("%f\n", (end_time - start_time) / 1000000000.0);
  }

  return NULL;
}

int main(int argc, char** argv) {
  // variables
  int i; // loop counter
  pthread_t* threads;
  pthread_attr_t attr;

  // parse arguments
  if (argc < 4) {
    printf("****************************************************\n");
    printf("* USAGE:                                           *\n");
    printf("* sa-seq <num_elems> <num_incr> <num_threads>      *\n");
    printf("* <num_elems>   - the size of the array            *\n");
    printf("* <num_incr>    - the number of increments to make *\n");
    printf("* <num_threads> - the number of threads to use     *\n");
    printf("****************************************************\n");
    exit(-1);
  } else {
    num_elems = atoi(argv[1]);
    num_incr = atoi(argv[2]);
    num_threads = atoi(argv[3]);
    /*
    printf("num_elems: %d\n", num_elems);
    printf("num_incr: %d\n", num_incr);
    printf("num_threads: %d\n", num_threads);
    */
    assert(num_incr % num_threads == 0);
  }

  // create and initialize shared array
  sa = (int*) malloc(num_elems * sizeof(int));
  for (i = 0; i < num_elems; i++) {
    sa[i] = 0;
  }
  //printf("*** INIT ***\n");
  //print_array(sa, num_elems);

  // initialize thread-related stuff
  threads = (pthread_t*) malloc(num_threads * sizeof(pthread_t));
  pthread_attr_init(&attr);
  pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);  
  pthread_mutex_init(&sa_lock, NULL);
  barrier_init(&bar, num_threads);

  //srand(time(0));
  srand(0x55555555);

  // start the threads in the thread_main routine
  for (i = 0; i < num_threads; i++) {
    assert(!pthread_create(&threads[i], &attr, increment, (void*) i));
  }

  // as each thread completes, merge with it
  for (i = 0; i < num_threads; i++) {
    assert(!pthread_join(threads[i], NULL));
  }

  //printf("*** FINAL ***\n");
  //print_array(sa, num_elems);
  verify(sa, num_elems, num_incr);

  // clean up thread-related stuff
  pthread_attr_destroy(&attr);
  pthread_mutex_destroy(&sa_lock);
  barrier_destroy(&bar);
  return 0;
}
